import numpy as np
from pyscipopt import Model, quicksum
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.tree import DecisionTreeRegressor
from utils import read_csv_to_dict, train_torch_neural_network

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of an agency tasked with replanting a stretch of land.

We have access to open source data (thanks to:
@article{wood2023tree,
  title={Tree seedling functional traits mediate plant-soil
  feedback survival responses across a gradient of light availability},
  author={Wood, Katherine EA and Kobe, Richard K and Ib{\'a}{\~n}ez, In{\'e}s and McCarthy-Neumann, Sarah},
  journal={Plos one},
  volume={18},
  number={11},
  pages={e0293906},
  year={2023},
  publisher={Public Library of Science San Francisco, CA USA}
}
Data accessed here: https://www.kaggle.com/datasets/yekenot/tree-survival-prediction)
for which we have built a predictor. The predictor takes as input a variety of features ranging
from soil properties to light properties, and determines the survival rate of the planted tree.

The goal of this MIP is to plant a set of trees such that some budget constraints are satisfied,
some diversity criteria on planted trees is satisfied, and the survival rate of the
planted trees is maximised. In the stretch of land where the trees are planted, the location
contains static features such as predicted sunlight, but the planter has the chance to sterilise sections of soil.
The planter must then decide in which location to plant which trees.

Let I, J be the index set of grid coordinates
Let T be the index set of the type of trees that can be planted
Let K be the index set of the features
Let x[i][j][k] be the value of feature k for the soil at grid point (i,j)
Let p[i][j][t] be the binary variable indicating iff tree t is planted at (i,j)
Let s[i][j]t[] be the survival rate of tree t if planted at grid point (i,j)
Let s'[i][j][t] be the adjusted survival rate (0 if not planted) of tree t at grid point (i,j)
Let min_survive[t] be the minimum amount of trees that need to survive of type t
Let costs[t] be the cost of planting tree t, and budget be the total budget
Let sterilisation_budget be the amount of grid points that be sterilised
Let f be the ML predictor for the survival of tree t at grid point i,j with soil feature x[i][j]

The MIP model is:

sum_t p[i][j][t] = 1 for all i,j
sum_{i,j} s'[i][j][t] >= min_survive[t] forall t
sum_t (sum_{i,j} p[i][j][t] * cost[t]) <= budget
sum_{i,j} x[i][j][sterilisation_idx] <= sterilisation_budget
s[i][j][t] = f(x[i][j][:])
p[i][j][t] = 1 => s'[i][j][t] <= s[i][j][t] forall i,j,t
p[i][j][t] = 0 => s'[i][j][t] <= 0

max(sum_{i,j,t} s'[i][j][t])
"""


def build_and_optimise_tree_planting(
    data_seed=42,
    training_seed=42,
    predictor_type="linear",
    formulation="sos",
    framework="sklearn",
    max_depth=5,
    n_estimators_layers=2,
    layer_size=8,
    n_grid_size=10,
    build_only=False,
):
    assert predictor_type in ("linear", "decision_tree", "gbdt", "mlp")
    training_random_state = np.random.RandomState(training_seed)
    data_random_state = np.random.RandomState(data_seed)
    max_sterilise = data_random_state.randint(low=8, high=13)
    costs = data_random_state.uniform(low=25, high=45, size=4)
    min_trees = data_random_state.uniform(
        low=n_grid_size**2 / 80, high=n_grid_size**2 / 60, size=4
    )
    max_budget = (n_grid_size**2) * np.median(costs) * 1.2
    # Read the csv data
    data_dict = read_csv_to_dict("./tests/data/tree_survivability.csv")

    # Extract the min and max light values
    light_values = np.array(
        [float(data_dict["Light_ISF"][i]) for i in range(len(data_dict["Light_ISF"]))]
    )
    min_light = np.min(light_values)
    max_light = np.max(light_values)

    # Now create the X and y data from the csv that is used to train the regression models
    species = sorted(list(set(data_dict["Species"])))
    species_idx_dict = {tree_type: i for i, tree_type in enumerate(species)}
    n_entries = len(data_dict["Light_ISF"])
    X = [[] for _ in range(len(species))]
    y = [[] for _ in range(len(species))]
    for i in range(n_entries):
        idx = species_idx_dict[data_dict["Species"][i]]
        X[idx].append([])
        if data_dict["Event"][i] == "1":
            y[idx] += [0]
        elif data_dict["Harvest"][i] == "X":
            y[idx] += [1]
        else:
            assert data_dict["Alive"][i] == "X", print(i)
            y[idx] += [1]
        X[idx][-1].append(light_values[i])
        X[idx][-1].append(data_dict["Light_Cat"][i] == "low")
        X[idx][-1].append(data_dict["Light_Cat"][i] == "med")
        X[idx][-1].append(data_dict["Light_Cat"][i] == "high")
        X[idx][-1].append(data_dict["Myco"][i] == "AMF")
        X[idx][-1].append(data_dict["SoilMyco"][i] == "AMF")
        X[idx][-1].append(data_dict["Sterile"][i] == "Sterile")

    # Create and train the regression model
    regression_models = []
    for i in range(len(species)):
        X_i = np.array(X[i])
        y_i = np.array(y[i])
        if predictor_type == "linear":
            reg = LinearRegression().fit(X_i, y_i)
        elif predictor_type == "decision_tree":
            reg = DecisionTreeRegressor(
                random_state=training_random_state, max_depth=max_depth
            ).fit(X_i, y_i)
        elif predictor_type == "gbdt":
            reg = GradientBoostingRegressor(
                random_state=training_random_state,
                n_estimators=n_estimators_layers,
                max_depth=max_depth,
            ).fit(X_i, y_i.reshape(-1))
        elif predictor_type == "mlp":
            if framework == "sklearn":
                hidden_layers = tuple([layer_size for i in range(n_estimators_layers)])
                reg = MLPRegressor(
                    random_state=training_random_state, hidden_layer_sizes=hidden_layers
                ).fit(X_i, y_i.reshape(-1))
            else:
                reg = train_torch_neural_network(
                    X_i, y_i, n_estimators_layers, layer_size, training_seed, reshape=True
                )
        else:
            raise ValueError(f"Unknown predictor type: {predictor_type}")
        regression_models.append(reg)

    # Initialise the SCIP Model
    scip = Model()

    # Create variables deciding which tree is planted in which grid point and the survival rate of each planted tree
    tree_planting_vars = np.zeros((n_grid_size, n_grid_size, 4), dtype=object)
    tree_survive_vars = np.zeros((n_grid_size, n_grid_size, 4), dtype=object)
    tree_adjusted_survive_vars = np.zeros((n_grid_size, n_grid_size, 4), dtype=object)
    for i in range(n_grid_size):
        for j in range(n_grid_size):
            for k in range(4):
                tree_planting_vars[i][j][k] = scip.addVar(
                    vtype="B", name=f"plant_tree_{i}_{j}_{k}"
                )
                tree_survive_vars[i][j][k] = scip.addVar(
                    vtype="C", lb=-10, ub=10, name=f"survive_tree_{i}_{j}_{k}"
                )
                tree_adjusted_survive_vars[i][j][k] = scip.addVar(
                    vtype="C", lb=-10, ub=1, name=f"adjusted_survive_tree_{i}_{j}_{k}"
                )

    # Ensure that only a single tree is planted in a single grid point
    for i in range(n_grid_size):
        for j in range(n_grid_size):
            scip.addCons(
                quicksum(tree_planting_vars[i][j][k] for k in range(4)) == 1,
                name=f"square_{i}_{j}",
            )

    # Ensure an expected minimum amount of each tree type survives
    for k in range(4):
        scip.addCons(
            quicksum(
                quicksum(tree_adjusted_survive_vars[i][j][k] for j in range(n_grid_size))
                for i in range(n_grid_size)
            )
            >= min_trees[k],
            name=f"min_trees_{k}",
        )

    # Ensure that the budget is respected, and not more trees are planted than can be afforded
    scip.addCons(
        quicksum(
            quicksum(
                quicksum(costs[k] * tree_planting_vars[i][j][k] for k in range(4))
                for j in range(n_grid_size)
            )
            for i in range(n_grid_size)
        )
        <= max_budget,
        name=f"tree_budget",
    )

    # Ensure that a tree can only survive if it is planted
    for i in range(n_grid_size):
        for j in range(n_grid_size):
            for k in range(4):
                scip.addConsIndicator(
                    tree_adjusted_survive_vars[i][j][k] <= tree_survive_vars[i][j][k],
                    tree_planting_vars[i][j][k],
                    name=f"tree_survive_ind_{i}_{j}_{k}_0",
                )
                scip.addConsIndicator(
                    tree_adjusted_survive_vars[i][j][k] <= 0,
                    tree_planting_vars[i][j][k],
                    activeone=False,
                    name=f"tree_survive_ind_{i}_{j}_{k}_2",
                )

    # Create feature variables
    feature_variables = np.zeros((n_grid_size, n_grid_size, 7), dtype=object)
    for i in range(n_grid_size):
        light_isf = max_light - ((i / n_grid_size) * (max_light - min_light))
        for j in range(n_grid_size):
            light_cat = data_random_state.randint(0, 3)
            myco = data_random_state.randint(0, 2)
            soil_myco = data_random_state.randint(0, 2)
            for k in range(7):
                if k == 0:
                    feature_variables[i][j][k] = scip.addVar(
                        vtype="C", lb=min_light, ub=max_light, name=f"feature_{i}_{j}_{k}"
                    )
                    scip.fixVar(feature_variables[i][j][k], light_isf)
                else:
                    feature_variables[i][j][k] = scip.addVar(
                        vtype="B", name=f"feature_{i}_{j}_{k}"
                    )
                    if (
                        (k == 1 and light_cat == 0)
                        or (k == 2 and light_cat == 1)
                        or (k == 3 and light_cat == 2)
                        or (k == 4 and myco == 0)
                        or (k == 5 and soil_myco == 0)
                    ):
                        scip.fixVar(feature_variables[i][j][k], 1)
                    elif k == 6:
                        pass
                    else:
                        scip.fixVar(feature_variables[i][j][k], 0)

    # Create a constraint that only a certain amount of squares can be sterilised
    scip.addCons(
        quicksum(
            quicksum(feature_variables[i][j][6] for j in range(n_grid_size))
            for i in range(n_grid_size)
        )
        <= max_sterilise,
        name="sterilisation_budget",
    )

    # Add the predictors to the MIP
    pred_cons_list = []
    for i in range(len(species)):
        pred_cons_list.append(
            add_predictor_constr(
                scip,
                regression_models[i],
                feature_variables.reshape(-1, 7),
                tree_survive_vars[:, :, i].reshape(-1, 1),
                unique_naming_prefix=f"predictor_{i}_",
                epsilon=0.0001,
                formulation=formulation,
            )
        )

    # Set the objective to maximise amount of trees that survive
    scip.setObjective(
        -quicksum(
            quicksum(
                quicksum(tree_adjusted_survive_vars[i][j][k] for k in range(4))
                for j in range(n_grid_size)
            )
            for i in range(n_grid_size)
        )
        + n_grid_size**2
    )

    if not build_only:
        # Optimise the model
        scip.optimize()

        # We can check the "error" of the MIP embedding via the difference between SKLearn and SCIP output
        for pred_cons in pred_cons_list:
            if np.max(pred_cons.get_error()) > 10**-5:
                error = np.max(pred_cons.get_error())
                raise AssertionError(f"Max error {error} exceeds threshold of {10 ** -5}")

    return scip


def test_tree_planting_linear():
    scip = build_and_optimise_tree_planting(
        data_seed=42,
        training_seed=42,
        predictor_type="linear",
        max_depth=5,
        n_estimators_layers=2,
        layer_size=8,
        n_grid_size=5,
    )


def test_tree_planting_decision_tree():
    scip = build_and_optimise_tree_planting(
        data_seed=42,
        training_seed=42,
        predictor_type="decision_tree",
        max_depth=5,
        n_estimators_layers=2,
        layer_size=8,
        n_grid_size=5,
    )


def test_tree_planing_gbdt():
    scip = build_and_optimise_tree_planting(
        data_seed=18,
        training_seed=35,
        predictor_type="gbdt",
        max_depth=3,
        n_estimators_layers=4,
        layer_size=4,
        n_grid_size=5,
    )


def test_tree_planting_mlp():
    scip = build_and_optimise_tree_planting(
        data_seed=18,
        training_seed=35,
        predictor_type="mlp",
        max_depth=5,
        n_estimators_layers=2,
        layer_size=4,
        n_grid_size=3,
    )


def test_tree_planting_mlp_bigm():
    scip = build_and_optimise_tree_planting(
        data_seed=18,
        training_seed=35,
        predictor_type="mlp",
        formulation="bigm",
        max_depth=5,
        n_estimators_layers=2,
        layer_size=4,
        n_grid_size=3,
    )



import numpy as np
from pyscipopt import Model
from sklearn.neural_network import MLPRegressor
from tensorflow import keras
from utils import train_torch_neural_network

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we are performing simple function approximations.
There are two non-linear functions that have been approximated by neural networks.
We want to minimise the value of the first function, while satisfying some equality of the second.

This is a simple example to test the performance of the MIP formulations for NNs
with ReLU activation functions:
@article{grimstad2019relu,
  title={ReLU networks as surrogate models in mixed-integer linear programs},
  author={Grimstad, Bjarne and Andersson, Henrik},
  journal={Computers and Chemical Engineering},
  volume={131},
  pages={106580},
  year={2019},
  publisher={Elsevier}
}

Let there be two random quadratic functions:
xQ_1x + a_1x + c_1
xQ_2x + a_2x + c_2

Let f be a NN that approximates the first quadratic function, and g be a NN
that approximates the second quadratic function.
Let const be some constant value in the range of the randomly generated quadratic function

The MIP model is:

g(x) = const
min(f(x))
"""


def build_and_optimise_function_approximation_model(
    n_inputs=5,
    n_samples=2500,
    framework="sklearn",
    formulation="sos",
    layer_size=8,
    n_layers=3,
    training_seed=42,
    data_seed=42,
    build_only=False,
):
    X, y_1, y_2 = build_random_quadratic_functions(
        seed=data_seed, n_inputs=n_inputs, n_samples=n_samples
    )

    if framework == "sklearn":
        hidden_layer_sizes = tuple([layer_size for _ in range(n_layers)])
        reg_1 = MLPRegressor(
            random_state=training_seed,
            hidden_layer_sizes=hidden_layer_sizes,
        ).fit(X, y_1.reshape(-1))
        reg_2 = MLPRegressor(
            random_state=training_seed,
            hidden_layer_sizes=hidden_layer_sizes,
        ).fit(X, y_2.reshape(-1))
    elif framework == "keras":
        keras.utils.set_random_seed(training_seed)
        reg_1 = keras.Sequential()
        reg_2 = keras.Sequential()
        reg_1.add(keras.Input(shape=(n_inputs,)))
        reg_2.add(keras.Input(shape=(n_inputs,)))
        for _ in range(n_layers):
            reg_1.add(keras.layers.Dense(layer_size, activation="relu"))
            reg_2.add(keras.layers.Dense(layer_size, activation="relu"))
        reg_1.add(keras.layers.Dense(1, activation="linear"))
        reg_2.add(keras.layers.Dense(1, activation="linear"))
        reg_1.compile(optimizer="adam", loss="mse")
        reg_1.fit(X, y_1, batch_size=64, epochs=20)
        reg_2.compile(optimizer="adam", loss="mse")
        reg_2.fit(X, y_2, batch_size=64, epochs=20)
    elif framework == "torch":
        reg_1 = train_torch_neural_network(
            X, y_1, n_layers, layer_size, training_seed, reshape=True
        )
        reg_2 = train_torch_neural_network(
            X, y_2, n_layers, layer_size, training_seed, reshape=True
        )
    else:
        raise ValueError(f"Framework {framework} is unknown")

    # Now build the SCIP Model and embed the neural networks
    scip, input_vars, output_vars = build_basic_scip_model(n_inputs, np.median(y_2))
    mlp_cons_1 = add_predictor_constr(
        scip,
        reg_1,
        input_vars,
        output_vars[0],
        unique_naming_prefix="reg_1_",
        formulation=formulation,
    )
    mlp_cons_2 = add_predictor_constr(
        scip,
        reg_2,
        input_vars,
        output_vars[1],
        unique_naming_prefix="reg_2_",
        formulation=formulation,
    )

    if not build_only:
        # Optimize the model
        scip.optimize()

        # We can check the "error" of the MIP embedding via the difference between SKLearn / Torch and SCIP output
        if np.max(mlp_cons_1.get_error()) > 2 * 10**-3:
            error = np.max(mlp_cons_1.get_error())
            raise AssertionError(f"Max error {error} exceeds threshold of {2 * 10 ** -3}")
        if np.max(mlp_cons_2.get_error()) > 2 * 10**-3:
            error = np.max(mlp_cons_2.get_error())
            raise AssertionError(f"Max error {error} exceeds threshold of {2 * 10 ** -3}")

    return scip


def build_random_quadratic_functions(seed=42, n_inputs=5, n_samples=1000):
    # Set the random seed so the results don't change
    np.random.seed(seed)

    # Generate two random quadratic functions f(x) = x^{t}Qx + Ax + c
    quadratic_coefficients = np.round(
        np.random.uniform(-5, 5, size=(2, n_inputs, n_inputs)), decimals=3
    )
    linear_coefficients = np.round(np.random.uniform(-1, 1, size=(2, n_inputs)), decimals=3)
    constant = np.round(np.random.uniform(-1, 1, size=(2,)), decimals=3)

    # Generate data from the quadratic function
    X = np.random.uniform(-10, 10, size=(n_samples, n_inputs))
    y_1 = np.zeros((n_samples,))
    y_2 = np.zeros((n_samples,))
    for i in range(n_samples):
        y_1[i] = (
            X[i] @ quadratic_coefficients[0] @ X[i].T
            + linear_coefficients[0] @ X[i].T
            + constant[0]
        )
        y_2[i] = (
            X[i] @ quadratic_coefficients[1] @ X[i].T
            + linear_coefficients[1] @ X[i].T
            + constant[1]
        )

    return X, y_1, y_2


def build_basic_scip_model(n_inputs, intercept):
    # Initialise a SCIP Model
    scip = Model()

    # Create the input variables
    input_vars = np.zeros((1, n_inputs), dtype=object)
    for i in range(n_inputs):
        # Tight bounds are important for MIP formulations of neural networks. They often drastically improve
        # performance. As our training data is in the range [-10, 10], we pass a multiple as bounds [-20, 20].
        # These bounds will then propagate to other variables.
        input_vars[0][i] = scip.addVar(name=f"x_{i}", vtype="C", lb=-20, ub=20)

    # Create the output variables. (Note that these variables will be automatically constructed if not specified)
    output_vars = np.zeros((2, 1), dtype=object)
    for i in range(2):
        output_vars[i] = scip.addVar(name=f"y_{i}", vtype="C", lb=None, ub=None)

    # Now set additional constraints and set the objective
    scip.addCons(output_vars[1][0] == intercept, name="fix_output_reg_2")
    scip.setObjective(output_vars[0][0] + 10000)

    return scip, input_vars, output_vars


def test_sklearn_mlp_regression():
    scip = build_and_optimise_function_approximation_model(
        data_seed=42,
        training_seed=42,
        n_inputs=5,
        n_samples=1000,
        framework="sklearn",
        formulation="sos",
        layer_size=8,
    )


def test_sklearn_mlp_regression_bigm():
    scip = build_and_optimise_function_approximation_model(
        data_seed=42,
        training_seed=42,
        n_inputs=5,
        n_samples=1000,
        framework="sklearn",
        formulation="bigm",
        layer_size=8,
    )


def test_torch_sequential_regression():
    scip = build_and_optimise_function_approximation_model(
        data_seed=42,
        training_seed=42,
        n_inputs=5,
        n_samples=1000,
        framework="torch",
        formulation="sos",
        layer_size=8,
    )


def test_torch_sequential_regression_bigm():
    scip = build_and_optimise_function_approximation_model(
        data_seed=42,
        training_seed=42,
        n_inputs=5,
        n_samples=1000,
        framework="torch",
        formulation="bigm",
        layer_size=8,
    )


def test_keras_sequential_regression():
    scip = build_and_optimise_function_approximation_model(
        data_seed=42,
        training_seed=42,
        n_inputs=5,
        n_samples=1000,
        framework="keras",
        formulation="sos",
        layer_size=3,
    )


def test_keras_sequential_regression_bigm():
    scip = build_and_optimise_function_approximation_model(
        data_seed=42,
        training_seed=42,
        n_inputs=5,
        n_samples=1000,
        framework="keras",
        formulation="bigm",
        layer_size=3,
    )



"""Module for formulating :external+torch:py:class:`torch.nn.Sequential` model in a PySCIPOPT Model.
"""

import numpy as np
import torch
from torch import nn

from ..exceptions import NoModel, NoSolution, ParameterError
from ..modelling.classification import argmax_bound_formulation
from ..modelling.neuralnet import BaseNNConstr


def add_sequential_constr(
    scip_model,
    sequential_model,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    output_type="regression",
    **kwargs,
):
    """Formulate sequential_model into scip_model.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the sequential model should be inserted.

    sequential_model : :external+torch:py:class:`torch.nn.Sequential`
        The sequential model to insert as predictor.

    input_vars : np.ndarray
        Decision variables used as input for the sequential neural network in PySCIPOpt Model.

    output_vars : np.ndarray
        Decision variables used as output for the sequential neural network in the PySCIPOpt Model

    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    output_type : {"classification", "regression"}, default="regression"
        If the option chosen is "classification" the output is 1 for exactly one class and
        0 for all others. If the option chosen is "regression" then the output for
        each node of the final layer is the value from the sequential torch model.

    Returns
    -------
    SequentialConstr
        Object containing information about what was added to model to insert the
        predictor into it

    Raises
    ------
    NoModel
        If a section of the Pytorch model structure (layer or activation) is not implemented.

    Warning
    -------
    Only :external+torch:py:class:`torch.nn.Linear`,
    :external+torch:py:class:`torch.nn.ReLU`,
    :external+torch:py:class:`torch.nn.Sigmoid`,
    :external+torch:py:class:`torch.nn.Tanh`,
    :external+torch:py:class:`torch.nn.Softmax`, and
    :external+torch:py:class:`torch.nn.Softplus` layers are supported.

    Note
    ----
    |VariablesDimensionsWarn|

    """
    return SequentialConstr(
        scip_model,
        sequential_model,
        input_vars,
        output_vars,
        unique_naming_prefix,
        output_type,
        **kwargs,
    )


class SequentialConstr(BaseNNConstr):
    """Transform a pytorch Sequential Neural Network to SCIP constraints with
    input and output as matrices of variables.

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        output_type="regression",
        **kwargs,
    ):
        if output_type not in ["regression", "classification"]:
            raise NoModel(
                predictor, f"Output type {output_type} is neither regression nor classification"
            )
        self.output_type = output_type
        out_features = None
        n_steps = len(predictor)
        for i, step in enumerate(predictor):
            if (
                output_type == "classification"
                and i == n_steps - 1
                and not isinstance(step, nn.Linear)
            ):
                continue
            if isinstance(step, nn.ReLU):
                pass
            elif isinstance(step, nn.Sigmoid):
                pass
            elif isinstance(step, nn.Tanh):
                pass
            elif isinstance(step, nn.Flatten):
                pass
            elif isinstance(step, nn.Softmax):
                pass
            elif isinstance(step, nn.Softplus):
                pass
            elif isinstance(step, nn.Linear):
                out_features = step.out_features
                pass
            else:
                raise NoModel(predictor, f"Unsupported layer {type(step).__name__}")
        if out_features is not None:
            self.output_size = out_features
        else:
            raise NoModel(predictor, "There is no Linear Layer in the given NN")
        super().__init__(
            scip_model, predictor, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Add the predictor constraints to SCIP."""

        # Get the input variables and the number of layers in the torch sequential object
        input_vars = self.input
        num_layers = len(self.predictor)
        output = self.output

        # Iterate through each step of the predictor and create the appropriate layers
        for i, step in enumerate(self.predictor):
            # In the final layer's case use the actual output instead of some created intermediary output
            if i == num_layers - 1:
                output = self.output

            # In the case of activation functions create the appropriate layer.
            # Ignore for classification on the final layer. All supported activation functions preserve the maximum
            if (
                isinstance(step, nn.ReLU)
                or isinstance(step, nn.Sigmoid)
                or isinstance(step, nn.Tanh)
                or isinstance(step, nn.Softmax)
                or isinstance(step, nn.Softplus)
            ) and (i < num_layers - 1 or self.output_type == "regression"):
                if isinstance(step, nn.ReLU):
                    activation = "relu"
                elif isinstance(step, nn.Sigmoid):
                    activation = "logistic"
                elif isinstance(step, nn.Softmax):
                    activation = "softmax"
                elif isinstance(step, nn.Softplus):
                    activation = "softplus"
                else:
                    activation = "tanh"
                if i < num_layers - 1:
                    output = self.scip_model.addMatrixVar(
                        input_vars.shape,
                        vtype="C",
                        lb=None,
                        ub=None,
                        name=self.unique_naming_prefix + f"layer_{i}",
                    )
                    self._created_vars.append(output)
                unique_naming_prefix = self.unique_naming_prefix + f"{activation}_{i}_"
                layer = self.add_activation_layer(
                    input_vars,
                    activation,
                    output,
                    unique_naming_prefix=unique_naming_prefix,
                    **kwargs,
                )
                input_vars = layer.output

            # In the case of a linear layer
            elif isinstance(step, nn.Linear):
                layer_weight, layer_bias = None, None
                for name, param in step.named_parameters():
                    if name == "weight":
                        layer_weight = param.detach().numpy().T
                    elif name == "bias":
                        layer_bias = param.detach().numpy()
                if layer_weight is None or layer_bias is None:
                    raise ParameterError(
                        "The torch sequential model linear layer contained no weights or bias!"
                    )
                if i < num_layers - 1 or self.output_type == "classification":
                    output = self.scip_model.addMatrixVar(
                        (input_vars.shape[0], layer_weight.shape[-1]),
                        vtype="C",
                        lb=None,
                        ub=None,
                        name=self.unique_naming_prefix + f"layer_{i}",
                    )
                    self._created_vars.append(output)
                layer = self.add_dense_layer(
                    input_vars,
                    layer_weight,
                    layer_bias,
                    "identity",
                    output,
                    unique_naming_prefix=self.unique_naming_prefix + f"linear_{i}_",
                    **kwargs,
                )
                input_vars = layer.output

            # In the case of a Flatten layer just do nothing.
            elif isinstance(step, nn.Flatten):
                pass

            # In the case of classification force the output to be binary with the argmax formulation
            if i == num_layers - 1 and self.output_type == "classification":
                if isinstance(step, nn.Sigmoid):
                    one_dim_center = 0
                else:
                    one_dim_center = 0.5
                new_vars, new_cons = argmax_bound_formulation(
                    self.scip_model,
                    input_vars,
                    self.output,
                    self.unique_naming_prefix,
                    one_dim_center=one_dim_center,
                )
                for var_set in new_vars:
                    self._created_vars.append(var_set)
                for cons_set in new_cons:
                    self._created_cons.append(cons_set)

    def get_error(self, eps=None):
        """
        Returns error in SCIP's solution with respect to the actual output of the sequential neural network

        Parameters
        ----------
        eps : float or int or None, optional
            The maximum allowed tolerance for a mismatch between the actual predictive model and SCIP.
            If the error is larger than eps an appropriate warning is printed

        Returns
        -------
        error: np.ndarray
            The absolute values of the difference between SCIP's solution and the trained ML model's output given
            the input as defined by SCIP. The matrix is the same dimension as the output of the trained predictor.
            Using torch / pyscipopt, the absolute difference between model.forward(input) and scip.getVal(output).

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """
        if self._has_solution:
            t_in = torch.from_numpy(self.input_values).float()
            t_out = self.predictor.forward(t_in)
            if self.output_type == "classification":
                t_out = nn.functional.one_hot(
                    torch.argmax(t_out, dim=1), num_classes=self.output_size
                )
            r_val = np.abs(t_out.detach().numpy() - self.output_values)
            if eps is not None and np.max(r_val) > eps:
                print(f"{t_out} != {self.output_values}")
            return r_val
        raise NoSolution()



from abc import ABC, abstractmethod

import numpy as np
from pyscipopt.scip import Constraint, MatrixVariable, Variable

from ..exceptions import NoSolution, ParameterError


class AbstractPredictorConstr(ABC):
    """Base class to store all information of embedded ML model by :py:func`pyscipopt_ml.add_predictor_constr`.

    This class is the base class to store everything that is added to
    a SCIP model when a trained predictor is inserted into it. Depending on
    the type of the predictor, a class derived from it will be returned
    by :py:func:`pyscipopt_ml.add_predictor_constr`.

    Warning
    -------

    Users should usually never construct objects of this class or one of its derived
    classes. They are returned by the :py:func:`pyscipopt_ml.add_predictor_constr` and
    other functions.
    """

    def __init__(
        self,
        scip_model,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        skip_validate=False,
        **kwargs,
    ):
        self.scip_model = scip_model
        self.unique_naming_prefix = unique_naming_prefix
        self._validate(input_vars, output_vars, skip_output=skip_validate)
        self._created_vars = []
        self._created_cons = []
        self._build_predictor_model(**kwargs)

    def _validate(self, input_vars, output_vars=None, skip_output=False):
        """Validate input and output variables (check shapes, reshape if needed)."""

        # Ensure the correct type of input and output is given
        if type(input_vars) not in [list, np.ndarray, MatrixVariable]:
            raise ParameterError(
                f"Input variables are not type list or np.ndarray. They are type {type(input_vars)}."
            )
        if output_vars is not None:
            if type(output_vars) not in [list, np.ndarray, MatrixVariable]:
                raise ParameterError(
                    f"Output variables are not type list or np.ndarray. They are type {type(output_vars)}."
                )

        # Transform the type list to type np.ndarray
        if isinstance(input_vars, list):
            input_vars = np.array(input_vars, dtype=object).view(MatrixVariable)
        elif not isinstance(input_vars, MatrixVariable):
            input_vars = input_vars.view(MatrixVariable)
        if isinstance(output_vars, list):
            output_vars = np.array(output_vars, dtype=object).view(MatrixVariable)
        elif output_vars is not None and not isinstance(output_vars, MatrixVariable):
            output_vars = output_vars.view(MatrixVariable)

        # Change the dimension of the input variables if needed. (Always want number of data points first)
        if input_vars.ndim == 1:
            input_vars = input_vars.reshape((1, -1))
        if input_vars.ndim >= 3:
            input_vars = input_vars.reshape((input_vars.shape[0], -1))

        # Ensure that the input variable dimensions match that of the predictor
        if (
            hasattr(self, "input_size")
            and self.input_size is not None
            and input_vars.shape[-1] != self.input_size
        ):
            raise ParameterError(
                f"Input variables dimension don't conform with predictor {type(self)} "
                + f"Input variable dimensions: {input_vars.shape[-1]} != {self.input_size}"
            )

        # Skip output validation if requested
        if not skip_output:
            # In the case of the output being None, create the appropriate output variables here
            if output_vars is None:
                output_vars = self._create_output_vars(input_vars)

            # Change the dimensions of the output variables if needed (Always want the number of data points first)
            if output_vars.ndim == 1:
                if input_vars.shape[0] == 1:
                    output_vars = output_vars.reshape((1, -1))
                else:
                    output_vars = output_vars.reshape((-1, 1))

            # Ensure that the output variable dimensions match that of the predictor
            if (
                hasattr(self, "output_size")
                and self.output_size is not None
                and output_vars.shape[-1] != self.output_size
            ):
                raise ParameterError(
                    f"Output variable dimensions don't conform with predictor {type(self)} "
                    + f"Output variable dimensions: {output_vars.shape[-1]} != {self.output_size}"
                )

            if output_vars.shape[0] != input_vars.shape[0]:
                raise ParameterError(
                    "Non-conforming dimension between input variables and output variables: "
                    + f"{output_vars.shape[0]} != {input_vars.shape[0]}"
                )

        self._input = input_vars
        self._output = output_vars

    def _build_predictor_model(self, **kwargs):
        self._mip_model(**kwargs)

    def _get_created_vars_and_cons(self):
        created_vars = [c_vars for c_vars in self._created_vars]
        created_cons = [c_cons for c_cons in self._created_cons]
        if hasattr(self, "_estimators"):
            for estimator in self._estimators:
                sub_created_vars, sub_created_cons = estimator._get_created_vars_and_cons()
                created_vars += sub_created_vars
                created_cons += sub_created_cons
        if hasattr(self, "_layers"):
            for layer in self._layers:
                sub_created_vars, sub_created_cons = layer._get_created_vars_and_cons()
                created_vars += sub_created_vars
                created_cons += sub_created_cons
        if hasattr(self, "_steps"):
            for step in self._steps:
                sub_created_vars, sub_created_cons = step._get_created_vars_and_cons()
                created_vars += sub_created_vars
                created_cons += sub_created_cons

        return created_vars, created_cons

    def print_stats(self, file=None):
        """Print statistics on model additions stored by this class.

        This function prints detailed statistics on the variables
        and constraints that were added to the model.

        Arguments
        ---------

        file: None, optional
            Text stream to which output should be redirected. By default, this is sys.stdout.
        """

        n_indicator_cons = 0
        n_sos_cons = 0
        n_linear_cons = 0
        n_nonlinear_cons = 0

        created_vars, created_cons = self._get_created_vars_and_cons()
        for cons_set in created_cons:
            it = np.nditer(cons_set, flags=["multi_index", "refs_ok"])
            for _ in it:
                if isinstance(cons_set[it.multi_index], Constraint):
                    cons_type = cons_set[it.multi_index].getConshdlrName()
                    if cons_type == "indicator":
                        n_indicator_cons += 1
                    elif cons_type == "SOS1":
                        n_sos_cons += 1
                    elif cons_type == "linear":
                        n_linear_cons += 1
                    elif cons_type == "nonlinear":
                        n_nonlinear_cons += 1
                    else:
                        raise TypeError(
                            f"Cons {cons_set[it.multi_index]} is of unknown type {cons_type}"
                        )

        n_bin_vars = 0
        n_cont_vars = 0

        for var_set in created_vars:
            it = np.nditer(var_set, flags=["multi_index", "refs_ok"])
            for _ in it:
                if isinstance(var_set[it.multi_index], Variable):
                    var_type = var_set[it.multi_index].vtype()
                    if var_type == "BINARY":
                        n_bin_vars += 1
                    elif var_type == "CONTINUOUS":
                        n_cont_vars += 1
                    else:
                        raise TypeError(
                            f"Var {var_set[it.multi_index]} is of unknown type {var_type}"
                        )

        print(
            f"Constraints created:\n Linear    {n_linear_cons}\n Indicator {n_indicator_cons}\n "
            f"SOS1      {n_sos_cons}\n Nonlinear {n_nonlinear_cons}\n"
            f"Created (internal) variables:\n Binary     {n_bin_vars}\n Continuous {n_cont_vars}\n"
            f"Input Shape:  {self.input.shape}\nOutput Shape: {self.output.shape}",
            file=file,
        )

    def _create_output_vars(self, input_vars):
        """May be defined in derived class to create the output variables of predictor."""
        if (not hasattr(self, "_output") or self._output is None) and (
            not hasattr(self, "output_size") or self.output_size is None
        ):
            raise AttributeError

        if not hasattr(self, "_output") or self._output is None:
            if hasattr(self, "classification"):
                if self.classification:
                    vtype = "B"
                else:
                    vtype = "C"
            else:
                vtype = "C"
            output_vars = self.scip_model.addMatrixVar(
                (input_vars.shape[0], self.output_size), vtype=vtype, lb=None, ub=None, name="out"
            )
            return output_vars
        else:
            return self._output

    @property
    def _has_solution(self):
        """Returns true if we have a solution."""
        if self.scip_model.getNSols() > 0:
            return True
        return False

    @abstractmethod
    def get_error(self, eps):
        """Returns error in SCIP's solution with respect to prediction from input.

        Returns
        -------
        error : ndarray of same shape as
            :py:attr:`pyscipopt_ml.modelling.base_predictor_constr.AbstractPredictorConstr.output`
            Assuming that we have a solution for the input and output variables
            `x, y`. Returns the absolute value of the differences between `predictor.predict(x)` and
            `y`. Where predictor is the regression / classification model represented by this object.

        Raises
        ------
        NoSolution
            If the SCIP model has no solution (either was not optimized or is infeasible).
        """
        ...

    @abstractmethod
    def _mip_model(self, **kwargs):
        """Makes MIP model for the predictor."""
        ...

    @property
    def input(self):
        """Returns the input variables of embedded predictor.

        Returns
        -------
        output : np.ndarray
        """
        return self._input

    @property
    def output(self):
        """Output variables of embedded predictor.

        Returns
        -------
        output : np.ndarray
        """
        return self._output

    @property
    def input_values(self):
        """Returns the values for the input variables if a solution is known.

        Returns
        -------
        input_vals : np.ndarray

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """
        if not self._has_solution:
            raise NoSolution

        input_vals = np.zeros(self.input.shape)
        for i in range(self.input.shape[0]):
            for j in range(self.input.shape[1]):
                input_vals[i][j] = self.scip_model.getVal(self.input[i][j])

        return input_vals

    @property
    def output_values(self):
        """Returns the values for the output variables if a solution is known.

        Returns
        -------
        output_value : np.ndarray

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """
        if not self._has_solution:
            raise NoSolution

        output_vals = np.zeros(self.output.shape)
        for i in range(self.output.shape[0]):
            for j in range(self.output.shape[1]):
                output_vals[i][j] = self.scip_model.getVal(self.output[i][j])

        return output_vals



import numpy as np
from pyscipopt import Model
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.neural_network import MLPClassifier
from tensorflow import keras
from utils import read_csv_to_dict, train_torch_neural_network

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of a health organisation.
This health organisation has a limited amount of supplies for treating water,
and they have the goal to treat as much water as possible such that the treated water
becomes drinkable. Regrettably, they lack resources to explicitly check the potability,
and must rely on some ML model that decides if the water is safe to drink or not.

We have access to open source data
(thanks to: https://github.com/MainakRepositor/Datasets/tree/master and
https://www.kaggle.com/datasets/uom190346a/water-quality-and-potability),
for which we can build predictors. The predictor takes as input a variety of features
about the water, and decides whether or not it is safe to drink.

The goal of this MIP is to treat the samples of water s.t. the largest amount
of water becomes drinkable. There are some constraints on which features can be treated,
due to to available resources.

Let I be the index set of the water samples
Let J be the index set of the features of the water
Let x[i][j] be the variable that takes the value for sample i and feature j after treatment
Let water_sample[i][j] be the value for sample i and feature j before treatment
Let f be the ML predictor that decides if x[i][:] is drinkable or not

The MIP formulation is:

x[i][j] = water_sample[i][j] - removed[i][j] + added[i][j] for all i, j
sum_i removed[i][j] <= remove_budget[j] for all j
sum_i added[i][j] <= added_budget[j] for all j
y[i] = f(x[i][:])

max(sum_i y[i])
"""


def build_and_optimise_water_potability(
    data_seed=42,
    training_seed=42,
    predictor_type="mlp",
    formulation="sos",
    n_water_samples=50,
    layer_size=16,
    max_depth=5,
    n_estimators_layers=3,
    framework="sklearn",
    build_only=False,
):
    assert predictor_type in ("mlp", "gbdt")
    # Random seed initialisation
    data_random_state = np.random.RandomState(data_seed)
    training_random_state = np.random.RandomState(training_seed)
    remove_feature_budgets = (
        data_random_state.uniform(1.8, 2.5),
        data_random_state.uniform(20, 25),
        data_random_state.uniform(180, 250),
        data_random_state.uniform(0.9, 1.1),
        data_random_state.uniform(19, 21),
        data_random_state.uniform(19, 21),
        data_random_state.uniform(1.2, 1.8),
        data_random_state.uniform(4.5, 6),
        data_random_state.uniform(0.9, 1.1),
    )
    add_feature_budgets = (
        data_random_state.uniform(1.8, 2.5),
        data_random_state.uniform(20, 25),
        data_random_state.uniform(180, 250),
        data_random_state.uniform(0.9, 1.1),
        data_random_state.uniform(19, 21),
        data_random_state.uniform(19, 21),
        data_random_state.uniform(1.2, 1.8),
        data_random_state.uniform(4.5, 6),
        data_random_state.uniform(0.9, 1.1),
    )

    # Path to water potability data
    data_dict = read_csv_to_dict("./tests/data/water_quality.csv")

    # The features of our predictor. All distance based features are variables.
    features = [
        "ph",
        "Hardness",
        "Solids",
        "Chloramines",
        "Sulfate",
        "Conductivity",
        "Organic_carbon",
        "Trihalomethanes",
        "Turbidity",
    ]
    n_features = len(features)

    # Generate the actual input data arrays for the ML predictors
    X = []
    y = np.array([int(x) for x in data_dict["Potability"]]).reshape(-1)
    for feature in features:
        X.append(np.array([float(x) for x in data_dict[feature]]))
    X = np.swapaxes(np.array(X), 0, 1)

    # Get the indices of some water that is currently undrinkable
    undrinkable_water_indices = (y == 0).nonzero()[0]
    np.random.shuffle(undrinkable_water_indices)
    undrinkable_water_indices = undrinkable_water_indices[:n_water_samples]

    # Build the MLP classifier
    if predictor_type == "gbdt":
        clf = GradientBoostingClassifier(
            random_state=training_random_state,
            max_depth=max_depth,
            n_estimators=n_estimators_layers,
        ).fit(X, y)
    else:
        if framework == "sklearn":
            hidden_layers = tuple([layer_size for i in range(n_estimators_layers)])
            clf = MLPClassifier(
                random_state=training_random_state,
                hidden_layer_sizes=hidden_layers,
            ).fit(X, y)
        elif framework == "keras":
            keras.utils.set_random_seed(training_seed)
            clf = keras.Sequential()
            clf.add(keras.Input(shape=(n_features,)))
            for i in range(n_estimators_layers):
                clf.add(keras.layers.Dense(layer_size, activation="relu"))
            clf.add(keras.layers.Dense(2, activation="linear"))
            clf.add(keras.layers.Activation(keras.activations.softmax))
            clf.compile(optimizer="adam", loss="binary_crossentropy")
            y_keras = np.zeros((len(y), 2))
            for i, class_val in enumerate(y):
                y_keras[i][class_val] = 1
            clf.fit(X, y_keras, batch_size=256, epochs=50)
        elif framework == "torch":
            clf = train_torch_neural_network(
                X,
                y,
                n_estimators_layers,
                layer_size,
                training_seed,
                reshape=True,
                binary_classifier=True,
            )
        else:
            raise ValueError(f"Unknown framework: {framework}")

    # Create the SCIP Model
    scip = Model()

    # Create variables deciding the features of each water sample after treatment
    feature_vars = np.zeros((n_water_samples, n_features), dtype=object)
    features_removed = np.zeros((n_water_samples, n_features), dtype=object)
    features_added = np.zeros((n_water_samples, n_features), dtype=object)
    if framework in ["sklearn", "torch"] or predictor_type == "gbdt":
        drinkable_water = np.zeros((n_water_samples, 1), dtype=object)
    else:
        drinkable_water = np.zeros((n_water_samples, 2), dtype=object)

    for i in range(n_water_samples):
        for j in range(drinkable_water.shape[-1]):
            drinkable_water[i][j] = scip.addVar(vtype="B", name=f"drinkable_{i}_{j}")
        for j in range(n_features):
            feature_vars[i][j] = scip.addVar(vtype="C", name=f"feature_val_{i}_{j}")
            features_added[i][j] = scip.addVar(
                vtype="C", lb=0, ub=add_feature_budgets[j], name=f"feature_add_{i}_{j}"
            )
            features_removed[i][j] = scip.addVar(
                vtype="C", lb=0, ub=remove_feature_budgets[j], name=f"feature_rem_{i}_{j}"
            )

    for i in range(n_water_samples):
        for j, feature in enumerate(features):
            lb = X[undrinkable_water_indices[i]][j] - remove_feature_budgets[j]
            ub = X[undrinkable_water_indices[i]][j] + add_feature_budgets[j]
            if lb >= 0:
                scip.chgVarLb(feature_vars[i][j], lb)
            scip.chgVarUb(feature_vars[i][j], ub)
            scip.addCons(
                X[undrinkable_water_indices[i]][j] - features_removed[i][j] + features_added[i][j]
                == feature_vars[i][j],
                name=f"change_{i}_{j}",
            )

    for j, feature in enumerate(features):
        scip.addCons(
            sum(features_added[i][j] for i in range(n_water_samples)) <= add_feature_budgets[j],
            name=f"add_budget_{feature}",
        )
        scip.addCons(
            sum(features_removed[i][j] for i in range(n_water_samples))
            <= remove_feature_budgets[j],
            name=f"remove_budget_{feature}",
        )

    # Add the ML predictor for predicting water quality of the sample
    if framework == "sklearn" or predictor_type == "gbdt":
        pred_cons = add_predictor_constr(
            scip,
            clf,
            feature_vars,
            drinkable_water,
            unique_naming_prefix=f"clf_",
            epsilon=0.0001,
            formulation=formulation,
        )
    else:
        pred_cons = add_predictor_constr(
            scip,
            clf,
            feature_vars,
            drinkable_water,
            unique_naming_prefix=f"clf_",
            output_type="classification",
            formulation=formulation,
        )

    # Set the object to maximise the amount of drinkable water after treatment
    if framework in ["sklearn", "torch"] or predictor_type == "gbdt":
        scip.setObjective(-np.sum(drinkable_water) + n_water_samples)
    else:
        scip.setObjective(-np.sum(drinkable_water[:, 1]) + n_water_samples)

    if not build_only:
        # Optimise the SCIP Model
        scip.optimize()

        # We can check the "error" of the MIP embedding via the difference between SKLearn and SCIP output
        if np.max(pred_cons.get_error()) > 10**-3:
            error = np.max(pred_cons.get_error())
            # TODO: There is currently no way to ensure SCIP numerical tolerances dont incorrectly classify

    return scip


def test_water_potability_sklearn_mlp():
    scip = build_and_optimise_water_potability(
        data_seed=42,
        training_seed=42,
        predictor_type="mlp",
        n_water_samples=50,
        layer_size=16,
        max_depth=5,
        n_estimators_layers=2,
        framework="sklearn",
        build_only=False,
    )


def test_water_potability_sklearn_mlp_bigm():
    scip = build_and_optimise_water_potability(
        data_seed=42,
        training_seed=42,
        predictor_type="mlp",
        formulation="bigm",
        n_water_samples=50,
        layer_size=16,
        max_depth=5,
        n_estimators_layers=2,
        framework="sklearn",
        build_only=False,
    )


def test_water_potability_keras():
    scip = build_and_optimise_water_potability(
        data_seed=20,
        training_seed=20,
        predictor_type="mlp",
        n_water_samples=50,
        layer_size=16,
        max_depth=5,
        n_estimators_layers=2,
        framework="keras",
        build_only=False,
    )


def test_water_potability_keras_bigm():
    scip = build_and_optimise_water_potability(
        data_seed=20,
        training_seed=20,
        predictor_type="mlp",
        formulation="bigm",
        n_water_samples=50,
        layer_size=16,
        max_depth=5,
        n_estimators_layers=2,
        framework="keras",
        build_only=False,
    )


def test_water_potability_gbdt():
    scip = build_and_optimise_water_potability(
        data_seed=18,
        training_seed=18,
        predictor_type="gbdt",
        n_water_samples=50,
        layer_size=16,
        max_depth=4,
        n_estimators_layers=4,
        framework="keras",
        build_only=False,
    )
